@@ -1,6 +1,6 @@
 #Copyright ReportLab Europe Ltd. 2000-2017
 #see license.txt for license details
-#history https://bitbucket.org/rptlab/reportlab/history-node/tip/src/reportlab/graphics/charts/barcharts.py
+#history https://hg.reportlab.com/hg-public/reportlab/log/tip/src/reportlab/graphics/charts/barcharts.py
 __version__='3.3.0'
 __doc__="""This module defines a variety of Bar Chart components.
 
@@ -10,12 +10,13 @@ vertical versions.
 """
 
 import copy, functools
+from ast import literal_eval
 
 from reportlab.lib import colors
 from reportlab.lib.validators import isNumber, isNumberOrNone, isColor, isColorOrNone, isString,\
             isListOfStrings, SequenceOf, isBoolean, isNoneOrShape, isStringOrNone,\
             NoneOr, isListOfNumbersOrNone, EitherOr, OneOf, isInt
-from reportlab.lib.utils import flatten
+from reportlab.lib.utils import flatten, isStr
 from reportlab.graphics.widgets.markers import uSymbol2Symbol, isSymbol
 from reportlab.lib.formatters import Formatter
 from reportlab.lib.attrmap import AttrMap, AttrMapValue
@@ -61,7 +62,7 @@ class BarChart(PlotArea):
         categoryAxis = AttrMapValue(None, desc='Handle of the category axis.'),
         data = AttrMapValue(None, desc='Data to be plotted, list of (lists of) numbers.'),
         barLabels = AttrMapValue(None, desc='Handle to the list of bar labels.'),
-        barLabelFormat = AttrMapValue(None, desc='Formatting string or function used for bar labels.'),
+        barLabelFormat = AttrMapValue(None, desc='Formatting string or function used for bar labels. Can be a list or list of lists of such.'),
         barLabelCallOut = AttrMapValue(None, desc='Callout function(label)\nlabel._callOutInfo = (self,g,rowNo,colNo,x,y,width,height,x00,y00,x0,y0)',advancedUsage=1),
         barLabelArray = AttrMapValue(None, desc='explicit array of bar label values, must match size of data if present.'),
         reversePlotOrder = AttrMapValue(isBoolean, desc='If true, reverse common category plot order.',advancedUsage=1),
@@ -200,8 +201,8 @@ class BarChart(PlotArea):
                 pdata = max((len(d) for d in D))*[0]
                 ndata = pdata[:]
                 for d in D:
-                    for i in xrange(len(d)):
-                        v = d[i] or 0
+                    for i,v in enumerate(d):
+                        v = v or 0
                         if v<=-1e-6:
                             ndata[i] += v
                         else:
@@ -231,8 +232,12 @@ class BarChart(PlotArea):
 
         # if zero is in chart, put the other axis there, otherwise use low
         crossesAt = vA.scale(0)
-        if crossesAt > org+length or crossesAt<org:
-            crossesAt = org
+        if not vA.forceZero:
+            eps = min(1,1.001*abs(length))
+            end = org+length
+            org, end = min(org,end)-eps, max(org,end)+eps
+            if crossesAt>eps or crossesAt<end:
+                crossesAt = org
         return crossesAt
 
     def _drawFinish(self):
@@ -277,7 +282,7 @@ class BarChart(PlotArea):
                 if k not in Z:
                     raise ValueError('Unknown zIndex variable %r in %r\nallowed variables are\n%s' % (k,Z,'\n'.join(['%s=%r'% (k,Z[k]) for k in sorted(Z.keys())])))
                 try:
-                    v = eval(v,{})  #only constants allowed
+                    v = literal_eval(v) #only constants allowed
                     assert isinstance(v,(float,int))
                 except:
                     raise ValueError('Bad zIndex value %r in clause %r of zIndex\nallowed variables are\n%s' % (v,z,zIndex,'\n'.join(['%s=%r'% (k,Z[k]) for k in sorted(Z.keys())])))
@@ -461,11 +466,10 @@ class BarChart(PlotArea):
             barRow.append(flipXY and (y,x,height,width) or (x,y,width,height))
 
         if style!='mixed':
-            for rowNo in xrange(seriesCount):   #iterate over the separate series
+            for rowNo, row in enumerate(data):  #iterate over the separate series
                 barRow = []
                 xVal = barsPerGroup - 1 - rowNo if reversePlotOrder else rowNo
                 xVal = offs + xVal*bGap
-                row = data[rowNo]
                 for colNo in xrange(rowLength): #iterate over categories
                     _addBar(colNo,colNo)
                 aBP(barRow)
@@ -484,11 +488,15 @@ class BarChart(PlotArea):
     def _getLabelText(self, rowNo, colNo):
         '''return formatted label text'''
         labelFmt = self.barLabelFormat
+        if isinstance(labelFmt,(list,tuple)):
+            labelFmt = labelFmt[rowNo]
+            if isinstance(labelFmt,(list,tuple)):
+                labelFmt = labelFmt[colNo]
         if labelFmt is None:
             labelText = None
         elif labelFmt == 'values':
             labelText = self.barLabelArray[rowNo][colNo]
-        elif type(labelFmt) is str:
+        elif isStr(labelFmt):
             labelText = labelFmt % self.data[rowNo][colNo]
         elif hasattr(labelFmt,'__call__'):
             labelText = labelFmt(self.data[rowNo][colNo])
@@ -602,7 +610,6 @@ class BarChart(PlotArea):
         g.add(r)
 
     def _makeBars(self,g,lg):
-        lenData = len(self.data)
         bars = self.bars
         br = getattr(self,'barRecord',None)
         BP = self._barPositions
@@ -612,11 +619,10 @@ class BarChart(PlotArea):
         catNNA = {}
         if catNAL:
             CBL = []
-            rowNoL = lenData - 1
+            rowNoL = len(self.data) - 1
             #find all the categories that have at least one value
             for rowNo, row in enumerate(BP):
-                for colNo in xrange(len(row)):
-                    x, y, width, height = row[colNo]
+                for colNo, (x, y, width, height) in enumerate(row):
                     if None not in (width,height):
                         catNNA[colNo] = 1
 
@@ -624,9 +630,8 @@ class BarChart(PlotArea):
             styleCount = len(bars)
             styleIdx = rowNo % styleCount
             rowStyle = bars[styleIdx]
-            for colNo in xrange(len(row)):
+            for colNo, (x,y,width,height) in enumerate(row):
                 style = (styleIdx,colNo) in bars and bars[(styleIdx,colNo)] or rowStyle
-                x, y, width, height = row[colNo]
                 if None in (width,height):
                     if not catNAL or colNo in catNNA:
                         self._addNABarLabel(lg,rowNo,colNo,x,y,width,height)
@@ -746,8 +751,8 @@ class BarChart(PlotArea):
             lo = self.x
             hi = lo + self.width
             end = self.y+self.height
-            for i in xrange(lenData):
-                for x, y, w, h in BP[i]:
+            for bp in BP:
+                for x, y, w, h in bp:
                     v = x+w
                     z = y+h
                     aC((min(y,z),max(y,z), min(x,v) - lo, hi - max(x,v)))
@@ -755,8 +760,8 @@ class BarChart(PlotArea):
             lo = self.y
             hi = lo + self.height
             end = self.x+self.width
-            for i in xrange(lenData):
-                for x, y, w, h in BP[i]:
+            for bp in BP:
+                for x, y, w, h in bp:
                     v = y+h
                     z = x+w
                     aC((min(x,z), max(x,z), min(y,v) - lo, hi - max(y,v)))
@@ -805,15 +810,12 @@ class BarChart(PlotArea):
         cA.configure(self._configureData)
         self.calcBarPositions()
 
-        lenData = len(self.data)
         bars = self.bars
         R = [].append
         BP = self._barPositions
-        for rowNo in xrange(lenData):
-            row = BP[rowNo]
+        for rowNo, row in enumerate(BP):
             C = [].append
-            for colNo in xrange(len(row)):
-                x, y, width, height = row[colNo]
+            for colNo, (x, y, width, height) in enumerate(row):
                 if None in (width,height):
                     na = self.naLabel
                     if na and na.text:
@@ -2355,3 +2357,4 @@ class SampleH5c4(Drawing):
         bc.categoryAxis.categoryNames = ['Ying', 'Yang']
 
         self.add(bc,name='HBC')
+        bc._computeSimpleBarLabelPositions()
